/* $Id$ */
/*-
 * Copyright (c) 2003, 2004 Benedikt Meurer <benny@xfce.org>
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <glib.h>

#include <libxfce4util/xdg.h>

#include <xprefs-daemon/xprefs-defaults.h>


static void   start_element_handler     (GMarkupParseContext *context,
                                         const gchar         *element_name,
                                         const gchar        **attribute_names,
                                         const gchar        **attribute_values,
                                         gpointer             user_data,
                                         GError             **error);
static void   end_element_handler       (GMarkupParseContext *context,
                                         const gchar         *element_name,
                                         gpointer             user_data,
                                         GError             **error);
static void   text_handler              (GMarkupParseContext *context,
                                         const gchar         *text,
                                         gsize                text_len,  
                                         gpointer             user_data,
                                         GError             **error);
static void   error_handler             (GMarkupParseContext *context,
                                         GError              *error,
                                         gpointer             user_data);


typedef enum _ParserState ParserState;
enum _ParserState
{
  PARSER_START,
  PARSER_DEFAULTS,
  PARSER_DEFAULT,
  PARSER_KEY,
  PARSER_VALUE,
};

typedef struct _Parser Parser;
struct _Parser
{
  ParserState       state;
  XPrefsDefaults   *defaults;
  const gchar      *filename;

  gchar            *key;
  XPrefsValue      *value;
  XPrefsValueType   value_type;
};


static void
start_element_handler (GMarkupParseContext *context,
                       const gchar         *element_name,
                       const gchar        **attribute_names,
                       const gchar        **attribute_values,
                       gpointer             user_data,
                       GError             **error)
{
  Parser *parser;
  int i;

  parser = (Parser *) user_data;

  switch (parser->state)
    {
    case PARSER_START:
      if (strcmp (element_name, "xprefs:defaults") == 0)
        parser->state = PARSER_DEFAULTS;
      break;

    case PARSER_DEFAULTS:
      if (strcmp (element_name, "xprefs:default") == 0)
        parser->state = PARSER_DEFAULT;
      break;

    case PARSER_DEFAULT:
      if (strcmp (element_name, "key") == 0)
        parser->state = PARSER_KEY;
      else if (strcmp (element_name, "value") == 0)
        {
          /* look for the type */
          parser->value_type = XPREFS_VALUE_INVALID;
          for (i = 0; attribute_names[i] != NULL; ++i)
            if (strcmp (attribute_names[i], "type") == 0)
              {
                parser->value_type =
                  xprefs_value_type_from_string (attribute_values[i]);
              }

          if (parser->value_type != XPREFS_VALUE_INVALID)
            parser->state = PARSER_VALUE;
        }
      break;

    default:
      /* XXX */
      break;
    }
}

static void
end_element_handler (GMarkupParseContext *context,
                     const gchar         *element_name,
                     gpointer             user_data,
                     GError             **error)
{
  Parser *parser;
  
  parser = (Parser *) user_data;

  switch (parser->state)
    {
    case PARSER_START:
      g_warning ("parser: This shouldn't happen.");
      break;

    case PARSER_DEFAULTS:
      if (strcmp (element_name, "xprefs:defaults") == 0)
        parser->state = PARSER_START;
      break;

    case PARSER_DEFAULT:
      if (strcmp (element_name, "xprefs:default") == 0)
        {
          if (parser->key != NULL)
            {
              xprefs_defaults_assign (parser->defaults,
                                      parser->key,
                                      parser->value);

              g_free (parser->key);
            }
          else if (parser->value != NULL)
            {
              xprefs_value_free (parser->value);
            }

          parser->key = NULL;
          parser->value = NULL;
          parser->state = PARSER_DEFAULTS;
        }
      break;

    case PARSER_KEY:
      if (strcmp (element_name, "key") == 0)
        parser->state = PARSER_DEFAULT;
      break;

    case PARSER_VALUE:
      if (strcmp (element_name, "value") == 0)
        parser->state = PARSER_DEFAULT;
      break;
    }
}

static void
text_handler (GMarkupParseContext *context,
              const gchar         *text,
              gsize                text_len,  
              gpointer             user_data,
              GError             **error)
{
  Parser *parser;
  
  parser = (Parser *) user_data;

  switch (parser->state)
    {
    case PARSER_KEY:
      if (parser->key != NULL)
        g_free (parser->key);
      parser->key = g_strdup (text);
      break;

    case PARSER_VALUE:
      if (parser->value != NULL)
        xprefs_value_free (parser->value);
      parser->value = xprefs_value_new_from_string (parser->value_type,
                                                    text,
                                                    error);
      break;

    default:
      /* XXX */
      break;
    }
}

static void
error_handler (GMarkupParseContext *context,
               GError              *error,
               gpointer             user_data)
{
  if (error != NULL && error->message != NULL)
    g_warning (" %s", error->message);
}


struct _XPrefsDefaults
{
  GHashTable *table;
};


/**
 * xprefs_defaults_new:
 *
 * Return value:
 **/
XPrefsDefaults*
xprefs_defaults_new (void)
{
  XPrefsDefaults *defaults;
  const gchar *filename;
  gchar **directories;
  gchar *path;
  GDir *dir;
  guint n;

  defaults        = g_new0 (XPrefsDefaults, 1);
  defaults->table = g_hash_table_new_full (g_str_hash,
                                           g_str_equal,
                                           (GDestroyNotify) g_free,
                                           (GDestroyNotify) xprefs_value_free);

  /* load defaults */
  directories = xfce_xdg_load_config_paths ("xprefs/defaults");
  for (n = 0; directories[n] != NULL; ++n)
    {
      if ((dir = g_dir_open (directories[n], 0, NULL)) != NULL)
        {
          while ((filename = g_dir_read_name (dir)) != NULL)
            if (g_str_has_suffix (filename, ".defaults"))
              {
                path = g_build_filename (directories[n], filename, NULL);
                xprefs_defaults_parse_file (defaults, path);
                g_free (path);
              }

          g_dir_close (dir);
        }
    }
  g_strfreev (directories);

  return defaults;
}

/**
 * xprefs_defaults_free:
 * @defaults:
 */
void
xprefs_defaults_free (XPrefsDefaults *defaults)
{
  g_return_if_fail (defaults != NULL);

  g_hash_table_destroy (defaults->table);
  g_free (defaults);
}


/**
 * xprefs_defaults_assign:
 * @defaults:
 * @key:
 * @value: new value for @key or NULL to remove any previous default value
 *         for @key. Note that @value will not be copied.
 *
 * Adds @value as default for @key to the @defaults overwriting any previous
 * default value for @key.
 **/
void
xprefs_defaults_assign (XPrefsDefaults *defaults,
                        const gchar *key,
                        XPrefsValue *value)
{
  g_return_if_fail (defaults != NULL);
  g_return_if_fail (key != NULL);

  if (value != NULL)
    {
      g_hash_table_replace (defaults->table,
                            g_strdup (key),
                            value);
    }
  else
    g_hash_table_remove (defaults->table, key);
}

/**
 * xprefs_defaults_lookup:
 * @defaults:
 * @key:
 *
 * Return value: the default value or NULL if theres no default for @key.
 **/
const XPrefsValue*
xprefs_defaults_lookup (const XPrefsDefaults *defaults,
                        const gchar *key)
{
  g_return_val_if_fail (defaults != NULL, NULL);
  g_return_val_if_fail (key != NULL, NULL);

  return (const XPrefsValue *) g_hash_table_lookup (defaults->table, key);
}


/**
 * xprefs_defaults_parse_file:
 * @defaults:
 * @filename:
 **/
void
xprefs_defaults_parse_file (XPrefsDefaults *defaults,
                            const gchar *filename)
{
  static GMarkupParser markup_parser = {
    start_element_handler,
    end_element_handler,
    text_handler,
    NULL,
    error_handler,
  };

  GMarkupParseContext *context;
  gchar *contents = NULL;
  Parser parser;

  if (!g_file_get_contents (filename, &contents, NULL, NULL))
    return;

  parser.state = PARSER_START;
  parser.defaults = defaults;
  parser.filename = filename;
  parser.key = NULL;
  parser.value = NULL;
  parser.value_type = XPREFS_VALUE_INVALID;

  context = g_markup_parse_context_new (&markup_parser, 0, &parser, NULL);
  g_markup_parse_context_parse (context, contents, strlen (contents), NULL);
  g_markup_parse_context_end_parse (context, NULL);
  g_markup_parse_context_free (context);

  g_free (contents);
}

