/* $Id$ */
/*-
 * Copyright (c) 2003, 2004 Benedikt Meurer <benny@xfce.org>
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif

#include <dbus/dbus-glib.h>

#include <xprefs/xprefs-dbus.h>
#include <xprefs/xprefs-engine.h>
#include <xprefs/xprefs-error.h>
#include <xprefs/xprefs-private.h>


typedef struct _ListenerGroup ListenerGroup;
typedef struct _Listener Listener;

struct _ListenerGroup
{
  gchar *namespace_section;
  GList *listeners;
};

struct _Listener
{
  guint               id;
  XPrefsListenerFunc  callback;
  gpointer            user_data;
};

#define LISTENER(listener)  ((Listener *) (listener))

struct _XPrefsEngine
{
  DBusConnection *connection;
  int             refcount;

  GHashTable     *listeners;
  GHashTable     *listener_ids;
};


static void
xprefs_engine_handle_notify (XPrefsEngine *engine,
                             DBusMessage  *message)
{
  ListenerGroup *group;
  DBusMessageIter iter;
  XPrefsEntry *entry;
  gchar *dir;

  dbus_message_iter_init (message, &iter);
  dir = dbus_message_iter_get_string (&iter);
  dbus_message_iter_next (&iter);
  entry = dbus_message_iter_get_xprefs_entry (&iter);

  group = g_hash_table_lookup (engine->listeners, dir);
  if (group != NULL)
    {
      GList *l;

      for (l = group->listeners; l != NULL; l = l->next)
        {
          Listener *listener = LISTENER (l->data);

          (*listener->callback) (engine,
                                 listener->id,
                                 entry,
                                 listener->user_data);
        }
    }

  xprefs_entry_free (entry);
  dbus_free (dir);
}

static void
xprefs_engine_handle_unregister (DBusConnection *connection,
                                 void *user_data)
{
}

static DBusHandlerResult
xprefs_engine_handle_message (DBusConnection *connection,
                              DBusMessage    *message,
                              void           *user_data)
{
  XPrefsEngine *engine = (XPrefsEngine *) user_data;

  if (dbus_message_is_method_call (message,
                                   XPREFS_DBUS_LISTENER_INTERFACE,
                                   XPREFS_DBUS_LISTENER_NOTIFY))
    {
      xprefs_engine_handle_notify (engine, message);
    }
  else
    {
      return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

  return DBUS_HANDLER_RESULT_HANDLED;
}

static const struct DBusObjectPathVTable xprefs_engine_handler_vtable =
{
  xprefs_engine_handle_unregister,
  xprefs_engine_handle_message,
  NULL,
};


XPrefsEngine*
xprefs_engine_new (DBusConnection *connection,
                   GError **error)
{
  XPrefsEngine *engine;
  gchar **path;

  g_return_val_if_fail (connection != NULL, NULL);

#if 0
  /* initialization */
  dbus_error_init (&dbus_error);

  /* try to activate the DataBase service */
  if (!dbus_bus_activate_service (dbus_conn,
                                  XPREFS_SERVICE_ORG_XFCE_XPREFS_DATABASE,
                                  0,
                                  NULL,
                                  &dbus_error))
    {
      dbus_set_g_error (error, &dbus_error);
      dbus_error_free (&dbus_error);

      return NULL;
    }
#endif

  /* create new connection object */
  engine              = g_new0 (XPrefsEngine, 1);
  engine->connection  = dbus_connection_ref (connection);
  engine->refcount    = 1;

  engine->listeners   = g_hash_table_new (g_str_hash, g_str_equal);
  engine->listener_ids= g_hash_table_new (g_str_hash, g_str_equal);

  /* register listener object */
  path = g_strsplit (XPREFS_DBUS_LISTENER_OBJECT + 1, "/", -1);
  dbus_connection_register_object_path (connection,
                                        (const gchar **) path,
                                        &xprefs_engine_handler_vtable,
                                        engine);
  g_free (path);

  return engine;
}

XPrefsEngine*
xprefs_engine_ref (XPrefsEngine *engine)
{
  g_return_val_if_fail (engine != NULL, NULL);
  engine->refcount++;
  return engine;
}

static void
free_listener_group (gpointer key, gpointer value, gpointer user_data)
{
  ListenerGroup *group = (ListenerGroup *) value;

  /* free the listeners in the group */
  g_list_foreach (group->listeners, (GFunc) g_free, NULL);
  g_list_free (group->listeners);

  /* free group and key */
  g_free (group);
  g_free (key);
}

/**
 * xprefs_engine_unref:
 **/
void
xprefs_engine_unref (XPrefsEngine *engine)
{
  g_return_if_fail (engine != NULL);

  if (--engine->refcount <= 0)
    {
      /* free listeners */
      g_hash_table_foreach (engine->listeners, free_listener_group, NULL);
      g_hash_table_destroy (engine->listener_ids);
      g_hash_table_destroy (engine->listeners);

      dbus_connection_unref (engine->connection);
      g_free (engine);
    }
}


static DBusMessage*
xprefs_engine_new_method_callv (const gchar *method,
                                int first_arg_type,
                                va_list valist)
{
  DBusMessage *message;

  /* create new message call */
  message = dbus_message_new_method_call (XPREFS_DBUS_SERVICE,
                                          XPREFS_DBUS_DATABASE_OBJECT,
                                          XPREFS_DBUS_DATABASE_INTERFACE,
                                          method);

  /* append initial arguments */
  dbus_message_append_args_valist (message, first_arg_type, valist);

  return message;
}

static DBusMessage*
xprefs_engine_new_method_call (const gchar *method,
                               int first_arg_type,
                               ...)
{
  DBusMessage *message;
  va_list valist;

  va_start (valist, first_arg_type);
  message = xprefs_engine_new_method_callv (method, first_arg_type, valist);
  va_end (valist);

  return message;
}

static void
xprefs_dbus_error_to_gerror (DBusError *derror,
                             GError **gerror)
{
  XPrefsError ecode;

  g_return_if_fail (dbus_error_is_set (derror));

  if (dbus_error_has_name (derror, XPREFS_DBUS_ERROR_FAILED))
    ecode = XPREFS_ERROR_FAILED;
  else if (dbus_error_has_name (derror, XPREFS_DBUS_ERROR_INVALID_VALUE))
    ecode = XPREFS_ERROR_INVALID_VALUE;
  else if (dbus_error_has_name (derror, XPREFS_DBUS_ERROR_PARSE_ERROR))
    ecode = XPREFS_ERROR_PARSE_ERROR;
  else if (dbus_error_has_name (derror, XPREFS_DBUS_ERROR_TYPE_MISMATCH))
    ecode = XPREFS_ERROR_TYPE_MISMATCH;
  else if (dbus_error_has_name (derror, XPREFS_DBUS_ERROR_BAD_KEY))
    ecode = XPREFS_ERROR_BAD_KEY;
  else if (dbus_error_has_name (derror, XPREFS_DBUS_ERROR_IS_DIR))
    ecode = XPREFS_ERROR_IS_DIR;
  else if (dbus_error_has_name (derror, XPREFS_DBUS_ERROR_IS_KEY))
    ecode = XPREFS_ERROR_IS_KEY;
  else
    {
      /* kind of D-BUS error */
      dbus_set_g_error (gerror, derror);
      dbus_error_free (derror);
      return;
    }

  g_set_error (gerror, XPREFS_ERROR, ecode, derror->message);
  dbus_error_free (derror);
}

static DBusMessage*
xprefs_engine_call (XPrefsEngine *engine,
                    GError **gerror,
                    const gchar *method,
                    int first_arg_type,
                    ...)
{
  DBusMessage *message;
  DBusMessage *reply;
  DBusError derror;
  va_list valist;

  /* prepare method call message */
  va_start (valist, first_arg_type);
  message = xprefs_engine_new_method_callv (method, first_arg_type, valist);
  va_end (valist);

  /* send method call message */
  dbus_error_init (&derror);
  reply = dbus_connection_send_and_process_reply (engine->connection,
                                                  message, &derror);
  dbus_message_unref (message);

  /* check error return */
  if (reply == NULL)
    {
      /* derror will be freed implicitly */
      xprefs_dbus_error_to_gerror (&derror, gerror);
    }

  return reply;
}

XPrefsValue*
xprefs_engine_get_value (XPrefsEngine *engine,
                         const gchar *key,
                         GError **gerror)
{
  DBusMessageIter iter;
  DBusMessage *reply;
  XPrefsValue *value = NULL;

  g_return_val_if_fail (engine != NULL, NULL);
  g_return_val_if_fail (key != NULL, NULL);

  /* send method call message */
  reply = xprefs_engine_call (engine, gerror,
                              XPREFS_DBUS_DATABASE_GET,
                              DBUS_TYPE_STRING, key,
                              DBUS_TYPE_UINT32, (dbus_uint32_t)0,
                              DBUS_TYPE_INVALID);

  /* process reply message */
  if (reply != NULL)
    {
      dbus_message_iter_init (reply, &iter); 
      value = dbus_message_iter_get_xprefs_value (&iter);
      dbus_message_unref (reply);

      if (value == NULL)
        {
          g_set_error (gerror, XPREFS_ERROR, XPREFS_ERROR_INVALID_VALUE,
                       _("Invalid value returned from Database object"));
        }
    }

  return value;
}

/**
 * xprefs_engine_set_value:
 * @engine:
 * @key:
 * @value:
 * @gerror:
 *
 * Return value: TRUE on success, else FALSE.
 **/
gboolean
xprefs_engine_set_value (XPrefsEngine *engine,
                         const gchar *key,
                         const XPrefsValue *value,
                         GError **gerror)
{
  DBusMessageIter iter;
  DBusMessage *message;
  DBusMessage *reply;
  DBusError derror;

  g_return_val_if_fail (engine != NULL, FALSE);
  g_return_val_if_fail (key != NULL, FALSE);
  g_return_val_if_fail (value != NULL, FALSE);

  /* prepare method call message */
  message = xprefs_engine_new_method_call (XPREFS_DBUS_DATABASE_SET,
                                           DBUS_TYPE_STRING, key,
                                           DBUS_TYPE_UINT32, (dbus_uint32_t)0,
                                           DBUS_TYPE_INVALID);
  dbus_message_append_iter_init (message, &iter);
  dbus_message_iter_append_xprefs_value (&iter, value);

  /* send message */
  dbus_error_init (&derror);
  reply = dbus_connection_send_and_process_reply (engine->connection,
                                                  message, &derror);
  dbus_message_unref (message);
  
  /* check for error reply */
  if (reply == NULL)
    {
      xprefs_dbus_error_to_gerror (&derror, gerror);
      return FALSE;
    }

  /* ... everything else means success */
  dbus_message_unref (reply);
  return TRUE;
}

/**
 * xprefs_engine_unset:
 * @connection: a XPrefsEngine.
 * @key: the key to unset.
 * @gerror: the return location for an allocated GError, or NULL to ignore
 *          errors.
 *
 * Unsets the value of @key; if @key is already unset, this operation has
 * no effect.
 *
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
xprefs_engine_unset (XPrefsEngine *engine,
                     const gchar *key,
                     GError **gerror)
{
  DBusMessage *reply;

  g_return_val_if_fail (engine != NULL, NULL);
  g_return_val_if_fail (key != NULL, NULL);

  /* send method call message */
  reply = xprefs_engine_call (engine, gerror,
                              XPREFS_DBUS_DATABASE_UNSET,
                              DBUS_TYPE_STRING, key,
                              DBUS_TYPE_UINT32, (dbus_uint32_t)0,
                              DBUS_TYPE_INVALID);

  /* if we get a method return reply, the unset operation worked */
  if (reply != NULL)
    {
      dbus_message_unref (reply);
      return TRUE;
    }

  /* ... else the operation failed and error is already set */
  return FALSE;
}

/**
 * xprefs_engine_dir_exists:
 * @engine: the XPrefsEngine object.
 * @key: the key to check.
 * @gerror: the return location for an allocated GError, or NULL to ignore
 *          errors.
 *
 * Return value: TRUE if the directory exists, FALSE if not or an error
 *               occured.
 **/
gboolean
xprefs_engine_dir_exists (XPrefsEngine *engine,
                          const gchar *key,
                          GError **gerror)
{
  DBusMessage *reply;
  dbus_bool_t result;
  DBusError derror;

  g_return_val_if_fail (engine != NULL, FALSE);
  g_return_val_if_fail (key != NULL, FALSE);

  result = FALSE;

  /* send method call message */
  reply = xprefs_engine_call (engine, gerror,
                              XPREFS_DBUS_DATABASE_DIR_EXISTS,
                              DBUS_TYPE_STRING, key,
                              DBUS_TYPE_UINT32, (dbus_uint32_t)0,
                              DBUS_TYPE_INVALID);

  if (reply != NULL)
    {
      dbus_error_init (&derror);
      if (!dbus_message_get_args (reply, &derror,
                                  DBUS_TYPE_BOOLEAN, &result,
                                  DBUS_TYPE_INVALID))
        {
          result = FALSE;
          dbus_set_g_error (gerror, &derror);
          dbus_error_free (&derror);
        }
      dbus_message_unref (reply);
    }

  return result;
}

/**
 * xprefs_engine_all_entries:
 * @engine: the XPrefsEngine object.
 * @key:
 * @gerror: the return location for an allocated GError, or NULL to ignore
 *          errors.
 *
 * Return value: the list of entries in @key.
 */
GList*
xprefs_engine_all_entries (XPrefsEngine *engine,
                           const gchar *key,
                           GError **gerror)
{
  DBusMessageIter iter;
  DBusMessage *reply;
  XPrefsEntry *entry;
  GList *entries;

  g_return_val_if_fail (engine != NULL, NULL);
  g_return_val_if_fail (key != NULL, NULL);

  entries = NULL;

  /* send method call message */
  reply = xprefs_engine_call (engine, gerror,
                              XPREFS_DBUS_DATABASE_ALL_ENTRIES,
                              DBUS_TYPE_STRING, key,
                              DBUS_TYPE_UINT32, (dbus_uint32_t)0,
                              DBUS_TYPE_INVALID);

  if (reply != NULL)
    {
      dbus_message_iter_init (reply, &iter);

      while (dbus_message_iter_has_next (&iter))
        {
          entry = dbus_message_iter_get_xprefs_entry (&iter);
          entries = g_list_append (entries, entry);
          dbus_message_iter_next (&iter);
        }

      dbus_message_unref (reply);
    }

  return entries;
}

/**
 * xprefs_engine_all_subdirs:
 * @engine: the XPrefsEngine object.
 * @key:
 * @gerror: the return location for an allocated GError, or NULL to ignore
 *          errors.
 *
 * Return value: the list of sub directory names in @key.
 */
GList*
xprefs_engine_all_subdirs (XPrefsEngine *engine,
                           const gchar *key,
                           GError **gerror)
{
  DBusMessageIter iter;
  DBusMessage *reply;
  GList *subdirs;
  gchar *subdir;

  g_return_val_if_fail (engine != NULL, NULL);
  g_return_val_if_fail (key != NULL, NULL);

  subdirs = NULL;

  /* send method call message */
  reply = xprefs_engine_call (engine, gerror,
                              XPREFS_DBUS_DATABASE_ALL_SUBDIRS,
                              DBUS_TYPE_STRING, key,
                              DBUS_TYPE_UINT32, (dbus_uint32_t)0,
                              DBUS_TYPE_INVALID);

  if (reply != NULL)
    {
      dbus_message_iter_init (reply, &iter);

      while (dbus_message_iter_has_next (&iter))
        {
          subdir = dbus_message_iter_get_string (&iter);
          subdirs = g_list_append (subdirs, subdir);
          dbus_free (subdir);
          dbus_message_iter_next (&iter);
        }

      dbus_message_unref (reply);
    }

  return subdirs;
}

static XPrefsValue*
xprefs_engine_get_type_checked_value (XPrefsEngine *engine,
                                      const gchar *key,
                                      XPrefsValueType type,
                                      GError **gerror)
{
  XPrefsValue *value;

  value = xprefs_engine_get_value (engine, key, gerror);
  if (value != NULL && G_UNLIKELY (value->type != type))
    {
      g_set_error (gerror, XPREFS_ERROR, XPREFS_ERROR_TYPE_MISMATCH,
                   _("Expected %s, got %s"),
                   xprefs_value_type_to_string (type),
                   xprefs_value_type_to_string (value->type));

      xprefs_value_free (value);
      value = NULL;
    }

  return value;
}

/**
 * xprefs_engine_get_bool:
 * @connection: a XPrefsEngine.
 * @key: key you want the value of.
 * @gerror: the return location for an allocated GError, or NULL to ignore
 *          errors.
 *
 * Requests the boolean value (XPREFS_VALUE_BOOL) stored at @key. Automatically
 * performs type-checking, so if a non-bool is stored at @key, an error is
 * returned. On error, or if @key is unset, FALSE is returned.
 *
 * Return value: the value of @key, or FALSE if no value is obtained.
 **/
gboolean
xprefs_engine_get_bool (XPrefsEngine *engine,
                        const gchar *key,
                        GError **gerror)
{
  XPrefsValue *value;
  gboolean data = FALSE;

  value = xprefs_engine_get_type_checked_value (engine, key,
                                                XPREFS_VALUE_BOOL,
                                                gerror);
  if (value != NULL)
    {
      data = xprefs_value_get_bool (value);
      xprefs_value_free (value);
    }

  return data;
}

/**
 * xprefs_engine_get_double:
 * @connection: a XPrefsEngine.
 * @key: key you want the value of.
 * @gerror: the return location for an allocated GError, or NULL to ignore
 *          errors.
 *
 * Requests the floating point number (XPREFS_VALUE_DOUBLE) stored at
 * @key. Automatically performs type-checking, so if a non-double value is
 * stored at @key, an error is returned. On error, or if @key is unset,
 * 0.0 is returned.
 *
 * Return value: the value of @key, or 0.0 if no value is obtained.
 **/
gdouble
xprefs_engine_get_double (XPrefsEngine *engine,
                          const gchar *key,
                          GError **gerror)
{
  XPrefsValue *value;
  gdouble data = 0.0;

  value = xprefs_engine_get_type_checked_value (engine, key,
                                                XPREFS_VALUE_DOUBLE,
                                                gerror);
  if (value != NULL)
    {
      data = xprefs_value_get_double (value);
      xprefs_value_free (value);
    }

  return data;
}

/**
 * xprefs_engine_get_int:
 * @connection: a XPrefsEngine.
 * @key: key you want the value of.
 * @gerror: the return location for an allocated GError, or NULL to ignore
 *          errors.
 *
 * Requests the integer (XPREFS_VALUE_INT) stored at @key. Automatically
 * performs type-checking, so if a non-int value is stored at @key, an error
 * is returned. On error, or if @key is unset, 0 is returned.
 *
 * Return value: the value of @key, or 0 if no value is obtained.
 **/
gint32
xprefs_engine_get_int (XPrefsEngine *engine,
                       const gchar *key,
                       GError **gerror)
{
  XPrefsValue *value;
  gint32 data = 0;

  value = xprefs_engine_get_type_checked_value (engine, key,
                                                XPREFS_VALUE_INT,
                                                gerror);
  if (value != NULL)
    {
      data = xprefs_value_get_int (value);
      xprefs_value_free (value);
    }

  return data;
}

/**
 * xprefs_engine_get_string:
 * @connection: a XPrefsEngine.
 * @key: key you want the value of.
 * @gerror: the return location for an allocated GError, or NULL to ignore
 *          errors.
 *
 * Requests the integer (XPREFS_VALUE_STRING) stored at @key. Automatically
 * performs type-checking, so if a non-string value is stored at @key, an error
 * is returned. On error, or if @key is unset, NULL is returned.
 *
 * Return value: the value of @key, or NULL if no value is obtained.
 **/
gchar*
xprefs_engine_get_string (XPrefsEngine *engine,
                          const gchar *key,
                          GError **gerror)
{
  XPrefsValue *value;
  gchar *data = NULL;

  value = xprefs_engine_get_type_checked_value (engine, key,
                                                XPREFS_VALUE_STRING,
                                                gerror);
  if (value != NULL)
    {
      data = xprefs_value_steal_string (value);
      xprefs_value_free (value);
    }

  return data;
}

static gboolean
xprefs_engine_set_type_checked_value (XPrefsEngine *engine,
                                      const gchar *key,
                                      XPrefsValue *value,
                                      GError **gerror)
{
  gboolean result;
  result = xprefs_engine_set_value (engine, key, value, gerror);
  xprefs_value_free (value);
  return result;
}

/**
 * xprefs_engine_set_bool:
 * @connection: a XPrefsEngine.
 * @key: key you want to set the value of.
 * @bool_data: new value of @key.
 * @gerror: the return location for an allocated GError, or NULL to ignore
 *          errors.
 *
 * Change the value of @key to @bool_data. Automatically creates the @key
 * if it didn't exists before (i.e. it was unset or it only had a default
 * value). If the key already exists but doesn't contain a boolean value
 * (XPREFS_VALUE_BOOL), xprefs_engine_set_bool() will fail.
 *
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
xprefs_engine_set_bool (XPrefsEngine *engine,
                        const gchar *key,
                        gboolean bool_data,
                        GError **gerror)
{
  XPrefsValue *value;

  g_return_val_if_fail (engine != NULL, FALSE);
  g_return_val_if_fail (key != NULL, FALSE);

  value = xprefs_value_new (XPREFS_VALUE_BOOL);
  xprefs_value_set_bool (value, bool_data);

  return xprefs_engine_set_type_checked_value (engine, key, value, gerror);
}

/**
 * xprefs_engine_set_double:
 * @connection: a XPrefsEngine.
 * @key: key you want to set the value of.
 * @double_data: new value of @key.
 * @gerror: the return location for an allocated GError, or NULL to ignore
 *          errors.
 *
 * Change the value of @key to @double_data. Automatically creates the @key
 * if it didn't exists before (i.e. it was unset or it only had a default
 * value). If the key already exists but doesn't contain a double value
 * (XPREFS_VALUE_DOUBLE), xprefs_engine_set_double() will fail.
 *
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
xprefs_engine_set_double (XPrefsEngine *engine,
                          const gchar *key,
                          gdouble double_data,
                          GError **gerror)
{
  XPrefsValue *value;

  g_return_val_if_fail (engine != NULL, FALSE);
  g_return_val_if_fail (key != NULL, FALSE);

  value = xprefs_value_new (XPREFS_VALUE_DOUBLE);
  xprefs_value_set_double (value, double_data);

  return xprefs_engine_set_type_checked_value (engine, key, value, gerror);
}


/**
 * xprefs_engine_set_int:
 * @connection: a XPrefsEngine.
 * @key: key you want to set the value of.
 * @int_data: new value of @key.
 * @gerror: the return location for an allocated GError, or NULL to ignore
 *          errors.
 *
 * Change the value of @key to @int_data. Automatically creates the @key
 * if it didn't exists before (i.e. it was unset or it only had a default
 * value). If the key already exists but doesn't contain an integer value
 * (XPREFS_VALUE_INT), xprefs_engine_set_int() will fail.
 *
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
xprefs_engine_set_int (XPrefsEngine *engine,
                       const gchar *key,
                       gint32 int_data,
                       GError **gerror)
{
  XPrefsValue *value;

  g_return_val_if_fail (engine != NULL, FALSE);
  g_return_val_if_fail (key != NULL, FALSE);

  value = xprefs_value_new (XPREFS_VALUE_INT);
  xprefs_value_set_int (value, int_data);

  return xprefs_engine_set_type_checked_value (engine, key, value, gerror);
}

/**
 * xprefs_engine_set_string:
 * @connection: a XPrefsEngine.
 * @key: key you want to set the value of.
 * @string_data: new value of @key.
 * @gerror: the return location for an allocated GError, or NULL to ignore
 *          errors.
 *
 * Change the value of @key to @string_data. Automatically creates the @key
 * if it didn't exists before (i.e. it was unset or it only had a default
 * value). If the key already exists but doesn't contain a string value
 * (XPREFS_VALUE_STRING), xprefs_engine_set_string() will fail.
 *
 * Return value: TRUE on success, FALSE on error.
 **/
gboolean
xprefs_engine_set_string (XPrefsEngine *engine,
                          const gchar *key,
                          const gchar *string_data,
                          GError **gerror)
{
  XPrefsValue *value;

  g_return_val_if_fail (engine != NULL, FALSE);
  g_return_val_if_fail (key != NULL, FALSE);

  value = xprefs_value_new (XPREFS_VALUE_STRING);
  xprefs_value_set_string (value, string_data);

  return xprefs_engine_set_type_checked_value (engine, key, value, gerror);
}


/**
 * xprefs_engine_listener_add:
 * @engine: a XPrefsEngine object to monitor for changes.
 * @namespace_section: the directory or key to watch; you'll be notified
 *                     of changes at or below this point.
 * @func: the callback to invoke when a notification is received from the
 *        database.
 * @user_data: the data to pass to the callback.
 * @gerror: the return location for an allocated GError, or NULL to ignore
 *          errors.
 *
 * Registers a listener request with the preferences database. The database
 * will notify the client when any key at or below @namespace_section
 * is set or unset. Try to watch the smallest possible part of the namespace
 * to not unnecessarily slow down the database and your application. The
 * function returns an ID you can use to remove the notification request; 0
 * is an invalid ID, and is returned in case of an error.
 *
 * Return value: an ID for the listener, or 0 on error.
 */
guint
xprefs_engine_listener_add (XPrefsEngine *engine,
                            const gchar *namespace_section,
                            XPrefsListenerFunc func,
                            gpointer user_data,
                            GError **gerror)
{
  ListenerGroup *group;
  Listener *listener;
  DBusMessage *reply;

  g_return_val_if_fail (engine != NULL, 0);
  g_return_val_if_fail (namespace_section != NULL, 0);

  group = g_hash_table_lookup (engine->listeners, namespace_section);
  if (group == NULL)
    {
      /* allocate the listener in the database */
      reply = xprefs_engine_call (engine, gerror,
                                  XPREFS_DBUS_DATABASE_LISTENER_ADD,
                                  DBUS_TYPE_STRING, namespace_section,
                                  DBUS_TYPE_INVALID);
      if (reply == NULL)
        return 0;
      dbus_message_unref (reply);

      group = g_new0 (ListenerGroup, 1);
      group->namespace_section = g_strdup (namespace_section);
      g_hash_table_insert (engine->listeners, group->namespace_section, group);
    }

  listener = g_new0 (Listener, 1);
  listener->id = GPOINTER_TO_UINT (listener);
  listener->callback = func;
  listener->user_data = user_data;

  group->listeners = g_list_prepend (group->listeners, listener);

  g_hash_table_insert (engine->listener_ids,
                       GUINT_TO_POINTER (listener->id),
                       group->namespace_section);

  return listener->id;
}

/**
 * xprefs_engine_listener_remove:
 * @engine: a XPrefsEngine object to monitor for changes.
 * @listener_id: the ID returned by xprefs_engine_listener_add().
 *
 * Removes a listener request.
 */
void
xprefs_engine_listener_remove (XPrefsEngine *engine,
                               guint listener_id)
{
  gchar *namespace_section;
  ListenerGroup *group;
  Listener *listener;
  GList *l;

  g_return_if_fail (engine != NULL);
  g_return_if_fail (listener_id > 0);

  /* lookup "namespace_section" by "listener_id" */
  namespace_section = g_hash_table_lookup (engine->listener_ids,
                                           GUINT_TO_POINTER (listener_id));
  if (namespace_section == NULL)
    return;

  /* remove "listener_id" -> "namespace_section" hash entry */
  g_hash_table_remove (engine->listener_ids, GUINT_TO_POINTER (listener_id));

  /* lookup "group" by "listener_id" */
  group = g_hash_table_lookup (engine->listeners,
                               namespace_section);
  if (group == NULL)
    return;

  for (l = group->listeners; l != NULL; l = l->next)
    if (((Listener *) l->data)->id == listener_id)
      break;

  if (l == NULL)
    return;

  listener = (Listener *) l->data;

  group->listeners = g_list_remove (group->listeners, listener);
  g_free (listener);

  if (g_list_length (group->listeners) == 0)
    {
      g_hash_table_remove (engine->listeners, group);
      g_free (group->namespace_section);
      g_free (group);
    }
}

